# Implementace haldy + příklady
#
# http://interactivepython.org/runestone/static/pythonds/Trees/BinaryHeapImplementation.html
# Jan Kybic, 2016


class MinHeap:
  """ binární halda, implementovaná pomocí metod """
  def __init__(self):
     self.heap = [] # indexujeme od nuly

  def bubble_up(self,i):
    """ probublá prvek 'i', zajistí splnění vlastnosti haldy """
    while i>0:
      j=(i-1)//2 # index rodiče  
      if self.heap[i] >= self.heap[j]:
        break
      self.heap[j],self.heap[i]=self.heap[i],self.heap[j]
      i = j
      
  def insert(self,k):
    self.heap+=[k]
    self.bubble_up(len(self.heap)-1)

  def peek(self):
    """ vrátí nejmenší prvek """
    return self.heap[0]

  def size(self):
    return len(self.heap)

  def is_empty(self):
    return self.size()==0 
  
  def bubble_down(self,i):
     n=self.size()
     while 2*i+1 < n:
        j=2*i+1 # zjisti index menšího syna
        if j+1 < n and self.heap[j] > self.heap[j+1]:
          j+=1
        if self.heap[i]>self.heap[j]:
          self.heap[i],self.heap[j]=self.heap[j],self.heap[i]
        i=j
                  
  def pop(self):
    """ odeber a vrať nejmenší prvek """
    element=self.heap[0]
    self.heap[0]=self.heap[-1]
    self.heap.pop() # smaž poslední prvek
    self.bubble_down(0)
    return element

class MaxHeap(MinHeap):

  def bubble_up(self,i):
    """ probublá prvek 'i', zajistí splnění vlastnosti haldy """
    while i>0:
      j=(i-1)//2 # index rodiče  
      if self.heap[i] <= self.heap[j]:
        break
      self.heap[j],self.heap[i]=self.heap[i],self.heap[j]
      i = j
  
  def bubble_down(self,i):
    n=self.size()
    while 2*i+1 < n:
      j=2*i+1 # zjisti index menšího syna
      if j+1 < n and self.heap[j] < self.heap[j+1]:
        j+=1
      if self.heap[i]<self.heap[j]:
        self.heap[i],self.heap[j]=self.heap[j],self.heap[i]
      i=j

  
def buildHeap(a,type=MinHeap):
    """ vytvoří haldu z obsahu pole, pole je přepsáno """
    h=type() # vytvoření MinHeap nebo MaxHeap
    h.heap=a # kopie
    for i in range((len(a)-1)//2,-1,-1):
        h.bubble_down(i)
    return h
  
# -----------------------------------------------------------------------------------        

def test_heap():
  h=MinHeap()
  for x in [9, 5, 0, 6, 2, 3]:
     h.insert(x)
  print(h.pop())
  h.insert(4)
  while not h.is_empty():
    print(h.pop(),end=", ")  
  print()
  h=buildHeap([9, 5, 0, 6, 2, 3])
  while not h.is_empty():
    print(h.pop(),end=", ")  
  print()
  h=buildHeap([9, 5, 0, 6, 2, 3],type=MaxHeap)
  while not h.is_empty():
    print(h.pop(),end=", ")  

     
if __name__=="__main__":
   test_heap()
